#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging

from modules.base.utils         import SimpleUtil as SU
from modules.base.exceptions    import NoActiveEntryException
# from modules.base.exceptions    import LQConnectionError, InvalidChannelException, NoActiveEntryException, EngineMalfunctionException, LQStreamException, LoadSourceException
from modules.base.mail          import AuraMailer
from modules.plugins.monitor    import AuraMonitor
from modules.core.state         import PlayerStateService


from modules.plugins.api import TrackserviceHandler


class EventBinding():
    """
    A binding between the event dispatcher and some event handler.

    This allows you to subscribe to events in this way:

        ```
        binding = dispatcher.attach(AuraMonitor)
        binding.subscribe("on_boot").subscribe("on_play")
        ```
    """
    dispatcher = None
    instance = None

    def __init__(self, dispatcher, instance):
        self.dispatcher = dispatcher
        self.instance = instance


    def subscribe(self, event_type):
        """
        Subscribes the instance to some event identified by the `event_type` string.
        """
        self.dispatcher.subscribe(self.instance, event_type)
        return self


    def get_instances(self):
        """
        Returns the object within that binding.
        """
        return self.instance



class EngineEventDispatcher():
    """
    Performs execution of handlers for engine events.
    """
    logger = None
    config = None

    subscriber_registry = None
    mailer = None
    soundsystem = None
    player_state = None
    scheduler = None
    api_handler = None


    def __init__(self, soundsystem, scheduler):
        """
        Initialize EventDispatcher
        """
        self.subscriber_registry = dict()
        self.logger = logging.getLogger("AuraEngine")
        self.config = soundsystem.config
        self.mailer = AuraMailer(self.config)
        self.soundsystem = soundsystem
        self.scheduler = scheduler
        self.player_state = PlayerStateService(self.config)

        # self.api_handler = ApiHandler(self.config)
        # self.monitoring = Monitoring(self.config, self.soundsystem)
        
        binding = self.attach(AuraMonitor)
        binding.subscribe("on_boot")

        binding = self.attach(TrackserviceHandler)
        binding.subscribe("on_play")



    def attach(self, clazz):
        """
        Creates an intance of the given Class.
        """
        instance = clazz(self.config, self.soundsystem)
        return EventBinding(self, instance)


    def subscribe(self, instance, event_type):
        """
        Subscribes to some event type. Preferably use it via `EventBinding.subscribe(..)`.
        """
        if not event_type in self.subscriber_registry:
            self.subscriber_registry[event_type] = []
        self.subscriber_registry[event_type].append(instance)


    def call_event(self, event_type, args):
        """
        Calls all subscribers for the given event type.
        """
        if not event_type in self.subscriber_registry:
            return
        listeners = self.subscriber_registry[event_type]
        if not listeners:
            return
        for listener in listeners:
            method = getattr(listener, event_type)
            if method:
                if args:
                    method(args)
                else:
                    method()


    #
    # Events
    #


    def on_initialized(self):
        """
        Called when the engine is initialized e.g. connected to Liquidsoap.
        """
        self.logger.debug("on_initialized(..)")
        self.scheduler.on_initialized()
        self.call_event("on_initialized", None)


    def on_boot(self):
        """
        Called when the engine is starting up. This happens after the initialization step.
        """
        self.logger.debug("on_boot(..)")
        self.call_event("on_boot", None)

        # self.monitoring.on_boot()


    def on_ready(self):
        """
        Called when the engine is booted and ready to play.
        """
        self.logger.debug("on_initialized(..)")
        self.scheduler.on_ready()


    def on_play(self, source):
        """
        Event Handler which is called by the soundsystem implementation (i.e. Liquidsoap)
        when some entry is actually playing.

        Args:
            source (String):    The URI of the media source currently being played
        """
        self.logger.debug("on_play(..)")

        self.logger.info(SU.pink("Source '%s' started playing" % source))
        try:
            self.player_state.store_trackservice_entry(source)
            # self.player_state.get_current_entry(source)
        except NoActiveEntryException:
            self.on_idle()

        self.call_event("on_initialized", None)


    def on_stop(self, entry):
        """
        The entry on the assigned channel has been stopped playing.
        """
        self.logger.debug("on_stop(..)")
        self.call_event("on_stop", entry)


    def on_idle(self):
        """
        Callend when no entry is playing
        """
        self.logger.debug("on_idle(..)")
        self.logger.warn(SU.red("Currently there's nothing playing!"))
        self.call_event("on_idle", None)


    def on_schedule_change(self, schedule):
        """
        Called when the playlist or entries of the current schedule have changed.
        """
        self.logger.debug("on_schedule_change(..)")
        self.call_event("on_schedule_change", schedule)


    def on_queue(self, entries):
        """
        One or more entries have been queued and are currently pre-loaded.
        """
        self.logger.debug("on_queue(..)")
        self.player_state.add_to_history(entries)
        self.call_event("on_queue", entries)


    def on_sick(self):
        """
        Called when the engine is in some unhealthy state.
        """
        self.logger.debug("on_sick(..)")
        self.call_event("on_sick", None)


    def on_resurrect(self):
        """
        Called when the engine turned healthy again after being sick.
        """
        self.logger.debug("on_resurrect(..)")
        self.call_event("on_resurrect", None)


    def on_critical(self, subject, message, data=None):
        """
        Callend when some critical event occurs
        """
        self.logger.debug("on_critical(..)")
        if not data: data = ""
        self.mailer.send_admin_mail(subject, message + "\n\n" + str(data))
        self.call_event("on_critical", (subject, message, data))
